/**
 * @file joystick.c
 *
 * @brief Handles joystick initialization.
 */

#include <string.h>
#include <SDL/SDL.h>
#include <SDL/SDL_joystick.h>
#if SDL_VERSION_ATLEAST(1,3,0)
#include "SDL/SDL_haptic.h"
#endif
#include "lephisto.h"
#include "log.h"
#include "joystick.h"

static SDL_Joystick* joystick = NULL; /**< Current joystick in use. */
#if SDL_VERSION_ATLEAST(1,3,0)
static int has_haptick = 0;     /**< Does the player have haptic? */
SDL_Haptic* haptic = NULL;      /**< Current haptic to use, externed in spfx.c */
unsigned int haptic_query = 0;  /**< Properties of the haptic device. */
#endif

static void joystick_initHaptic(void);

/**
 * @brief Get the joystick index by name.
 *    @param namjoystick Looks for given string in the joystick name.
 *    @return The index if found, defaults to 0 if it isn't found.
 */
int joystick_get(char* namjoystick) {
  int i;
  for(i = 0; i < SDL_NumJoysticks(); i++)
    if(strstr(SDL_JoystickName(i), namjoystick))
      return i;

  WARN("Joystick '%s' not found, using default joystick '%s'",
       namjoystick, SDL_JoystickName(0));
  return 0;
}

/**
 * @brief Make the game use a joystick by index.
 *    @param indjoystick Index of the joystick to use.
 *    @return 0 on success.
 */
int joystick_use(int indjoystick) {
  /* Check to see if it exists. */
  if(indjoystick < 0 || indjoystick >= SDL_NumJoysticks()) {
    WARN("Joystick of index number %d does not exist. Switching to default (0)",
         indjoystick);
    indjoystick = 0;
  }
  /* Close if already open. */
  if(joystick != NULL) {
    SDL_JoystickClose(joystick);
    joystick = NULL;
  }
  /* Start using the joystick. */
  joystick = SDL_JoystickOpen(indjoystick);
  if(joystick == NULL) {
    WARN("Error opening joystick %d [%s]",
         indjoystick, SDL_JoystickName(indjoystick));
    return -1;
  }
  LOG("Using joystick %d - %s", indjoystick, SDL_JoystickName(indjoystick));
  DEBUG("   With %d axes, %d buttons, %d balls, and %d hats\n",
        SDL_JoystickNumAxes(joystick), SDL_JoystickNumButtons(joystick),
        SDL_JoystickNumBalls(joystick), SDL_JoystickNumHats(joystick));

  /* Initialize the haptic if possible. */
  joystick_initHaptic();

  /* For style purposes. */
  DEBUG();

  return 0;
}

/**
 * @brief Initializes force feedback for the loaded device.
 */
static void joystick_initHaptic(void) {
#if SDL_VERSION_ATLEAST(1,3,0)
  if(has_haptic && SDL_JoystickIsHaptic(joystick)) {
    /* Close haptic if already open. */
    if(haptic != NULL) {
      SDL_HapticClose(haptic);
      haptic = NULL;
    }

    /* Try to create haptic device. */
    haptic = SDL_HapticOpenFromJoystick(joystick);
    if(haptic == NULL) {
      WARN("Unable to initialize force feedback.");
      return;
    }

    /* Check to see what it supports. */
    haptic_query = SDL_HapticQuery(haptic);
    if(!(haptic_query & SDL_HAPTIC_SINE)) {
      SDL_HapticClose(haptic);
      haptic = NULL;
      return;
    }
    DEBUG("   force feedback enabled");
  }
#endif
}

/**
 * @brief Initializes the joystick subsystem.
 *    @return 0 on success.
 */
int joystick_init(void) {
  int numjoysticks, i;

  /* Init the SDL subsys. */
  if(SDL_InitSubSystem(SDL_INIT_JOYSTICK)) {
    WARN("Unable to init the joystick subsystem.");
    return -1;
  }

#if SDL_VERSION_ATLEAST(1,3,0)
  if(SDL_InitSubSystem(SDL_INIT_HAPTIC) == 0)
    has_haptic = 1;
#endif

  /* Figure out how many joysticks there are. */
  numjoysticks = SDL_NumJoysticks();
  LOG("%d joystick%s detected", numjoysticks, (numjoysticks==1)?"":"s");
  for(i = 0; i < numjoysticks; i++)
    DEBUG("   %d. %s", i, SDL_JoystickName(i));

  /* Enable joystick events. */
  SDL_JoystickEventState(SDL_ENABLE);

  return 0;
}

/**
 * @brief Exit the joystick subsystem.
 */
void joystick_exit(void) {
#if SDL_VERSION_ATLEAST(1,3,0)
  if(haptic != NULL) {
    SDL_HapticClose(haptic);
    haptic = NULL;
  }
#endif
  if(joystick != NULL) {
    SDL_JoystickClose(joystick);
    joystick = NULL;
  }
}

