/**
 * @file save.c
 *
 * @brief Handles saving/loading games.
 */

#include <stdio.h> /* remove() */
#include <errno.h>

#include "lephisto.h"
#include "log.h"
#include "lxml.h"
#include "player.h"
#include "toolkit.h"
#include "dialogue.h"
#include "menu.h"
#include "lfile.h"
#include "hook.h"
#include "ldata.h"
#include "save.h"

#define LOAD_WIDTH    400
#define LOAD_HEIGHT   300

#define BUTTON_WIDTH  50
#define BUTTON_HEIGHT 30

/* Externs. */
/* player.c */
extern int  player_save(xmlTextWriterPtr writer); /**< Save player related stuff. */
extern int  player_load(xmlNodePtr parent); /**< Load player related stuff. */
/* mission.c */
extern int  missions_saveActive(xmlTextWriterPtr writer); /**< Save active missions. */
extern int  missions_loadActive(xmlNodePtr parent); /**< Load active missions. */
/* llua_misn.c */
extern int  var_save(xmlTextWriterPtr writer); /**< Save mission variables. */
extern int  var_load(xmlNodePtr parent); /**< Load mission variables. */
/* faction.c */
extern int  pfaction_save(xmlTextWriterPtr writer); /**< Save faction data. */
extern int  pfaction_load(xmlNodePtr parent); /**< Load faction data. */
/* hook.c */
extern int  hook_save(xmlTextWriterPtr writer); /**< Save hooks. */
extern int  hook_load(xmlNodePtr parent); /**< Load hooks. */
/* space.c */
extern int  space_sysSave(xmlTextWriterPtr writer); /**< Save the space stuff. */
extern int  space_sysLoad(xmlNodePtr parent); /**< Load the space stuff. */
/* unidiff.c */
extern int  diff_save(xmlTextWriterPtr writer); /**< Save the universe diffs. */
extern int  diff_load(xmlNodePtr parent); /**< Load the universe diffs. */
/* menu.c */
extern void menu_main_close(void); /**< Close the main menu. */

/* Static. */
static int  save_data(xmlTextWriterPtr writer);
static void load_menu_close(unsigned int wdw, char* str);
static void load_menu_load(unsigned int wdw, char* str);
static void load_menu_delete(unsigned wdw, char* str);
static int  load_game(char* file);

/**
 * @brief Save all the players game data.
 *    @param writer XML writer to use.
 *    @return 0 on success.
 */
static int save_data(xmlTextWriterPtr writer) {
  /* The data itself. */
  if(diff_save(writer) < 0)             return -1; /* Must save first or can get cleared. */
  if(player_save(writer) < 0)           return -1;
  if(missions_saveActive(writer) < 0)   return -1;
  if(var_save(writer) < 0)              return -1;
  if(pfaction_save(writer) < 0)         return -1;
  if(hook_save(writer) < 0)             return -1;
  if(space_sysSave(writer) < 0)         return -1;
  return 0;
}

/**
 * @brief Save the current game.
 *    @return 0 on success.
 */
int save_all(void) {
  char file[PATH_MAX], old[PATH_MAX];
  xmlDocPtr doc;
  xmlTextWriterPtr writer;

  /* Create the writer. */
  writer = xmlNewTextWriterDoc(&doc, 1);
  if(writer == NULL) {
    ERR("testXmlwriterDoc: Error creating the xml writer.");
    return -1;
  }

  /* Set the writer parameters. */
  xmlTextWriterSetIndentString(writer, (const xmlChar*)" ");
  xmlTextWriterSetIndent(writer, 1);

  /* Start element. */
  xmlw_start(writer);
  xmlw_startElem(writer, "lephisto_save");
  
  /* Save the version or something.. */
  xmlw_startElem(writer, "version");
  xmlw_elem(writer, "lephisto", "%d.%d.%d", VMAJOR, VMINOR, VREV);
  xmlw_elem(writer, "data", ldata_name());
  xmlw_endElem(writer); /* Version. */
  
  /* Save the data. */
  if(save_data(writer) < 0) {
    ERR("Trying to save game data");
    goto err_writer;
  }

  /* Finish the element. */
  xmlw_endElem(writer); /* lephisto_save. */
  xmlw_done(writer);

  /* Write to file. */
  if(lfile_dirMakeExist("%ssaves", lfile_basePath()) < 0) {
    WARN("Aborting save...");
    goto err_writer;
  }
  snprintf(file, PATH_MAX, "%ssaves/%s.ls", lfile_basePath(), player_name);

  /* Back up old savegame. */
  snprintf(old, PATH_MAX, "%s.backup", file);
  if(lfile_fileExists(file)) {
    if(rename(file, old) < 0) {
      WARN("Unable to back up old save: %s. Not saving!", strerror(errno));
      goto err_writer;
    }
  }

  /* Critical section, if crashes, the players save game gets screwed!!!
   * Luckily we have a copy just in case..
   */
  xmlFreeTextWriter(writer);
  if(xmlSaveFileEnc(file, doc, "UTF-8") < 0) {
    WARN("Failed to write savegame! You'll most likey have to restore it by \
          copying your backup savegame over your current savegame.");
    goto err;
  }
  xmlFreeDoc(doc);

  return 0;

err_writer:
  xmlFreeTextWriter(writer);
err:
  xmlFreeDoc(doc);
  return -1;
}

/**
 * @brief Open the load game menu.
 */
void load_game_menu(void) {
  unsigned int wid;
  char** files;
  int lfiles, i, len;

  /* Window. */
  wid = window_create("Load Game", -1, -1, LOAD_WIDTH, LOAD_HEIGHT);

  /* Load the saves. */
  files = lfile_readDir(&lfiles, "%ssaves", lfile_basePath());
  for(i = 0; i < lfiles; i++) {
    len = strlen(files[i]);

    /* No save extension. */
    if((len < 6) || strcmp(&files[i][len-3], ".ls")) {
      free(files[i]);
      memmove(&files[i], &files[i+1], sizeof(char*) * (lfiles-i-1));
      lfiles--;
      i--;
    }
    else /* Remove the extension. */
      files[i][len-3] = '\0';
  }

  /* Again.. What if there is no files? */
  if(files == NULL) {
    files = malloc(sizeof(char*));
    files[0] = strdup("None");
    lfiles = 1;
  }

  window_addList(wid, 20, -50,
      LOAD_WIDTH-BUTTON_WIDTH-50, LOAD_HEIGHT-90,
      "lstSaves", files, lfiles, 0, NULL);

  /* Buttons. */
  window_addButton(wid, -20, 20, BUTTON_WIDTH, BUTTON_HEIGHT,
      "btnBack", "Back", load_menu_close);

  window_addButton(wid, -20, 30 + BUTTON_HEIGHT, BUTTON_WIDTH, BUTTON_HEIGHT,
      "btnLoad", "Load", load_menu_load);

  window_addButton(wid, -20, 20+2*(10+BUTTON_HEIGHT), BUTTON_WIDTH, BUTTON_HEIGHT,
      "btnDelete", "Del", load_menu_delete);

  /* Default action. */
  window_setAccept(wid, load_menu_load);
}

static void load_menu_close(unsigned int wdw, char* str) {
  (void)str;

  window_destroy(wdw);
}

static void load_menu_load(unsigned int wdw, char* str) {
  (void)str;
  char* save, path[PATH_MAX];
  int wid;

  wid = window_get("Load Game");

  save = toolkit_getList(wid, "lstSaves");

  if(strcmp(save, "None")==0)
    return;

  snprintf(path, PATH_MAX, "%ssaves/%s.ls", lfile_basePath(), save);

  /* Close menus before loading for proper rendering. */
  load_menu_close(wdw, NULL);
  menu_main_close();

  load_game(path);
}

/**
 * @brief Delete an old game.
 *    @param wdw Window to delete.
 *    @param str Unused.
 */
static void load_menu_delete(unsigned int wdw, char* str) {
  (void)str;
  char* save, path[PATH_MAX];
  int wid;

  wid = window_get("Load Game");
  save = toolkit_getList(wid, "lstSaves");

  if(strcmp(save, "None") == 0)
    return;

  if(dialogue_YesNo("Permanently Delete?",
        "Are you sure you want to permanently delete '%s'?", save) == 0)
    return;

  snprintf(path, PATH_MAX, "%ssaves/%s.ls", lfile_basePath(), save);
  remove(path); /* Remove is more portable and will call unlink on Linux. */

  /* Need to reload the menu. */
  load_menu_close(wdw, NULL);
  load_game_menu();
}

/* Load a new game. */
static int load_game(char* file) {
  xmlNodePtr node;
  xmlDocPtr doc;

  doc = xmlParseFile(file);
  node = doc->xmlChildrenNode; /* Base node. */
  if(node == NULL) {
    WARN("Savegame '%s' invalid!", file);
    return -1;
  }

  diff_load(node);    /* Must load first to work properly. */
  player_load(node);
  var_load(node);
  missions_loadActive(node);
  pfaction_load(node);
  hook_load(node);
  space_sysLoad(node);

  /* Initialize the economy. */
  economy_init();

  /* Need to run takeoff hooks since player just "took off". */
  hooks_run("takeoff");
  hooks_run("enter");

  xmlFreeDoc(doc);
  xmlCleanupParser();

  return 0;
}

